/*
 * %W% %E%
 *
 * Copyright (c) 2006, Oracle and/or its affiliates. All rights reserved.
 * ORACLE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package java.util.concurrent.atomic;
                                                                                //IBM-Oracle_security_bulletin
import java.lang.reflect.Array;                                                 //IBM-Oracle_security_bulletin
import java.util.Arrays;                                                        //IBM-Oracle_security_bulletin
import sun.misc.Unsafe;

/**
 * An array of object references in which elements may be updated
 * atomically.  See the {@link java.util.concurrent.atomic} package
 * specification for description of the properties of atomic
 * variables.
 * @since 1.5
 * @author Doug Lea
 * @param <E> The base class of elements held in this array
 */
public class AtomicReferenceArray<E> implements java.io.Serializable {
    private static final long serialVersionUID = -6209656149925076980L;

    private static final Unsafe unsafe;                                         //IBM-Oracle_security_bulletin
    private static final int base;                                              //IBM-Oracle_security_bulletin
    private static final int shift;                                             //IBM-Oracle_security_bulletin
    private static final long arrayFieldOffset;                                 //IBM-Oracle_security_bulletin
    private final Object[] array; // must have exact type Object[]              //IBM-Oracle_security_bulletin

    static {                                                                    //IBM-Oracle_security_bulletin
        int scale;                                                              //IBM-Oracle_security_bulletin
        try {                                                                   //IBM-Oracle_security_bulletin
            unsafe = Unsafe.getUnsafe();                                        //IBM-Oracle_security_bulletin
            arrayFieldOffset = unsafe.objectFieldOffset                         //IBM-Oracle_security_bulletin
                (AtomicReferenceArray.class.getDeclaredField("array"));         //IBM-Oracle_security_bulletin
            base = unsafe.arrayBaseOffset(Object[].class);                      //IBM-Oracle_security_bulletin
            scale = unsafe.arrayIndexScale(Object[].class);                     //IBM-Oracle_security_bulletin
        } catch (Exception e) {                                                 //IBM-Oracle_security_bulletin
            throw new Error(e);                                                 //IBM-Oracle_security_bulletin
        }                                                                       //IBM-Oracle_security_bulletin
        if ((scale & (scale - 1)) != 0)                                         //IBM-Oracle_security_bulletin
            throw new Error("data type scale not a power of two");              //IBM-Oracle_security_bulletin
        shift = 31 - Integer.numberOfLeadingZeros(scale);                       //IBM-Oracle_security_bulletin
    }                                                                           //IBM-Oracle_security_bulletin
                                                                                //IBM-Oracle_security_bulletin
    private long checkedByteOffset(int i) {                                     //IBM-Oracle_security_bulletin
        if (i < 0 || i >= array.length)
            throw new IndexOutOfBoundsException("index " + i);
        return byteOffset(i);                                                   //IBM-Oracle_security_bulletin
    }                                                                           //IBM-Oracle_security_bulletin
                                                                                //IBM-Oracle_security_bulletin
    private static long byteOffset(int i) {                                     //IBM-Oracle_security_bulletin
        return ((long) i << shift) + base;                                      //IBM-Oracle_security_bulletin
    }

    /**
     * Creates a new AtomicReferenceArray of given length.
     * @param length the length of the array
     */
    public AtomicReferenceArray(int length) {
        array = new Object[length];
    }

    /**
     * Creates a new AtomicReferenceArray with the same length as, and
     * all elements copied from, the given array.
     *
     * @param array the array to copy elements from
     * @throws NullPointerException if array is null
     */
    public AtomicReferenceArray(E[] array) {
        // Visibility guaranteed by final field guarantees                      //IBM-Oracle_security_bulletin
        this.array = Arrays.copyOf(array, array.length, Object[].class);        //IBM-Oracle_security_bulletin
    }

    /**
     * Returns the length of the array.
     *
     * @return the length of the array
     */
    public final int length() {
        return array.length;
    }

    /**
     * Gets the current value at position {@code i}.
     *
     * @param i the index
     * @return the current value
     */
    public final E get(int i) {
        return getRaw(checkedByteOffset(i));                                    //IBM-Oracle_security_bulletin
    }                                                                           //IBM-Oracle_security_bulletin
                                                                                //IBM-Oracle_security_bulletin
    private E getRaw(long offset) {                                             //IBM-Oracle_security_bulletin
        return (E) unsafe.getObjectVolatile(array, offset);                     //IBM-Oracle_security_bulletin
    }

    /**
     * Sets the element at position {@code i} to the given value.
     *
     * @param i the index
     * @param newValue the new value
     */
    public final void set(int i, E newValue) {
        unsafe.putObjectVolatile(array, checkedByteOffset(i), newValue);        //IBM-Oracle_security_bulletin
    }

    /**
     * Eventually sets the element at position {@code i} to the given value.
     *
     * @param i the index
     * @param newValue the new value
     * @since 1.6
     */
    public final void lazySet(int i, E newValue) {
        unsafe.putOrderedObject(array, checkedByteOffset(i), newValue);         //IBM-Oracle_security_bulletin
    }


    /**
     * Atomically sets the element at position {@code i} to the given
     * value and returns the old value.
     *
     * @param i the index
     * @param newValue the new value
     * @return the previous value
     */
    public final E getAndSet(int i, E newValue) {
        long offset = checkedByteOffset(i);                                     //IBM-Oracle_security_bulletin
        while (true) {
            E current = getRaw(offset);                                         //IBM-Oracle_security_bulletin
            if (compareAndSetRaw(offset, current, newValue))                    //IBM-Oracle_security_bulletin
                return current;
        }
    }

    /**
     * Atomically sets the element at position {@code i} to the given
     * updated value if the current value {@code ==} the expected value.
     *                                                                          //IBM-Oracle_security_bulletin
     * @param i the index
     * @param expect the expected value
     * @param update the new value
     * @return true if successful. False return indicates that
     * the actual value was not equal to the expected value.
     */
    public final boolean compareAndSet(int i, E expect, E update) {
        return compareAndSetRaw(checkedByteOffset(i), expect, update);          //IBM-Oracle_security_bulletin
    }                                                                           //IBM-Oracle_security_bulletin
                                                                                //IBM-Oracle_security_bulletin
    private boolean compareAndSetRaw(long offset, E expect, E update) {         //IBM-Oracle_security_bulletin
        return unsafe.compareAndSwapObject(array, offset, expect, update);      //IBM-Oracle_security_bulletin
    }

    /**
     * Atomically sets the element at position {@code i} to the given
     * updated value if the current value {@code ==} the expected value.
     *
     * <p>May <a href="package-summary.html#Spurious">fail spuriously</a>
     * and does not provide ordering guarantees, so is only rarely an
     * appropriate alternative to {@code compareAndSet}.
     *
     * @param i the index
     * @param expect the expected value
     * @param update the new value
     * @return true if successful.
     */
    public final boolean weakCompareAndSet(int i, E expect, E update) {
        return compareAndSet(i, expect, update);
    }

    /**
     * Returns the String representation of the current values of array.
     * @return the String representation of the current values of array.
     */
    public String toString() {
        int iMax = array.length - 1;                                            //IBM-Oracle_security_bulletin
        if (iMax == -1)                                                         //IBM-Oracle_security_bulletin
            return "[]";                                                        //IBM-Oracle_security_bulletin
                                                                                //IBM-Oracle_security_bulletin
        StringBuilder b = new StringBuilder();                                  //IBM-Oracle_security_bulletin
        b.append('[');                                                          //IBM-Oracle_security_bulletin
        for (int i = 0; ; i++) {                                                //IBM-Oracle_security_bulletin
            b.append(getRaw(byteOffset(i)));                                    //IBM-Oracle_security_bulletin
            if (i == iMax)                                                      //IBM-Oracle_security_bulletin
                return b.append(']').toString();                                //IBM-Oracle_security_bulletin
            b.append(',').append(' ');                                          //IBM-Oracle_security_bulletin
        }                                                                       //IBM-Oracle_security_bulletin
    }                                                                           //IBM-Oracle_security_bulletin
                                                                                //IBM-Oracle_security_bulletin
    /**                                                                         //IBM-Oracle_security_bulletin
     * Reconstitutes the instance from a stream (that is, deserializes it).     //IBM-Oracle_security_bulletin
     * @param s the stream                                                      //IBM-Oracle_security_bulletin
     */                                                                         //IBM-Oracle_security_bulletin
    private void readObject(java.io.ObjectInputStream s)                        //IBM-Oracle_security_bulletin
        throws java.io.IOException, ClassNotFoundException {                    //IBM-Oracle_security_bulletin
        // Note: This must be changed if any additional fields are defined      //IBM-Oracle_security_bulletin
        Object a = s.readFields().get("array", null);                           //IBM-Oracle_security_bulletin
        if (a == null || !a.getClass().isArray())                               //IBM-Oracle_security_bulletin
            throw new java.io.InvalidObjectException("Not array type");         //IBM-Oracle_security_bulletin
        if (a.getClass() != Object[].class)                                     //IBM-Oracle_security_bulletin
            a = Arrays.copyOf((Object[])a, Array.getLength(a), Object[].class); //IBM-Oracle_security_bulletin
        unsafe.putObjectVolatile(this, arrayFieldOffset, a);                    //IBM-Oracle_security_bulletin
    }

}
//IBM-atomic_concurrent
//IBM-Oracle_security_bulletin
